#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Material.h"
#include "MaterialPoint.h"

/* Warnings */
static const Result_t WRN_NO_TEX = {
	RESULT_WARNING, "The material has no attached texture. Weird shading may follow"};

static const Result_t WRN_NO_LIGHT = {
	RESULT_WARNING, "There is no light source specified for Geometry node"};

/*---------------------------------------------------------------------------*/

/* Shader prefix */
static char *prefix = \
	"#define POINT\n"
;

/* Shaders */
static const char *vshader = "../shaders/MaterialPhong.vs";
static const char *fshader = "../shaders/MaterialPhong.fs";

/*---------------------------------------------------------------------------*/

MaterialPoint::MaterialPoint(
		const char		*nidentifier,
		u32				purge_level
	) : Material(nidentifier, purge_level, MATERIAL_POINT)
{
	shaderProgram = sceneGraph.createShaderProgram(nidentifier, prefix,
		vshader, fshader, purge_level);

	diffuseTexture = 0;

	ambient = rgbf(0.0f, 0.1f, 0.2f);
	diffuse = rgbf(0.1f, 0.5f, 0.7f);
	specular = rgbf(0.7f, 0.9f, 1.0f);
	specularLevel = 1.0f;
	shininess = 3.4f;
	emissive = 0.0f;

	Texture::setTexture(&diffuseTexture, "../textures/notex.png", NO_PURGE);
}

/*---------------------------------------------------------------------------*/

MaterialPoint::~MaterialPoint()
{
	ensure_heap();

	if (diffuseTexture)
		sceneGraph.dereferenceResource(diffuseTexture);

	ensure_heap();
}

/*---------------------------------------------------------------------------*/

void MaterialPoint::prepareMaterial(
		FrameInfo_t			*di,
		Geometry			*geometry
	)
{
	u32 i;
	vec3f lpos;
	rgbf lcol;

	for (i = 0; i < MAX_SHADER_LIGHT_SLOTS; i++) {
		Light *light = geometry->getLight(i);

		if (light) {
			lpos.x = light->world.T0;
			lpos.y = light->world.T1;
			lpos.z = light->world.T2;
			lcol = light->getColor() * light->getIntensity();
			break;
		}
	}

	if (i == MAX_SHADER_LIGHT_SLOTS) {
		REPORT(WRN_NO_LIGHT);
		lpos = vec3f(0.0f, 0.0f, 0.0f);
		lcol = rgbf(1.0f, 1.0f, 1.0f);
	}

	shaderProgram->setUniformVector("LightColor", lcol.s, 1, 3);
	shaderProgram->setUniformVector("LightPosition", lpos.s, 1, 3);

	shaderProgram->setUniformVector("AmbientColor", ambient.s, 1, 3);
	shaderProgram->setUniformVector("DiffuseColor", diffuse.s, 1, 3);
	shaderProgram->setUniformVector("SpecularColor", specular.s, 1, 3);
	shaderProgram->setUniform("Shininess", &shininess, 1);
	shaderProgram->setUniform("Emissive", &emissive, 1);
	shaderProgram->setUniform("SpecularLevel", &specularLevel, 1);

	if (diffuseTexture) {
		shaderProgram->setTextureSampler("DiffuseSampler", diffuseTexture, 0);
	} else {
		REPORT(WRN_NO_TEX);
	}

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void MaterialPoint::setDiffuseTexture(
		Texture			*tex,

		bool			instantiate_resource
	)
{
	Texture::setTexture(&diffuseTexture, tex, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void MaterialPoint::setDiffuseTexture(
		const char		*filename,

		u32				purge_level
	)
{
	Texture::setTexture(&diffuseTexture, filename, purge_level);
}

/*---------------------------------------------------------------------------*/

Texture *MaterialPoint::getDiffuseTexture()
{
	return diffuseTexture;
}

/*---------------------------------------------------------------------------*/

Resource *MaterialPoint::clone(
		const char		*clone_name,

		u32				purge_level
	)
{
	MaterialPoint *new_res = (MaterialPoint *) sceneGraph.createMaterial(materialType,
		clone_name, purge_level);

	new_res->ambient = ambient;
	new_res->diffuse = diffuse;
	new_res->emissive = emissive;
	new_res->shininess = shininess;
	new_res->specular = specular;
	new_res->specularLevel = specularLevel;

	new_res->setDiffuseTexture(diffuseTexture, true);

	return new_res;
}

/*---------------------------------------------------------------------------*/

