#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Material.h"
#include "MaterialWeird.h"

/* Warnings */
static const Result_t WRN_NO_TEX = {
	RESULT_WARNING, "The material has no attached texture. Weird shading may follow"};

static const Result_t WRN_NO_LIGHT = {
	RESULT_WARNING, "There is no light source specified for Geometry node"};

/*---------------------------------------------------------------------------*/

/* Shader prefix */
static char *prefix = \
	"\n"
;

/* Shaders */
static const char *vshader = "../shaders/MaterialWeird.vs";
static const char *fshader = "../shaders/MaterialWeird.fs";

/*---------------------------------------------------------------------------*/

MaterialWeird::MaterialWeird(
		const char		*nidentifier,
		u32				purge_level
	) : Material(nidentifier, purge_level, MATERIAL_WEIRD)
{
	shaderProgram = sceneGraph.createShaderProgram(nidentifier, prefix,
		vshader, fshader, purge_level);

	diffuseTexture = 0;
	noiseTexture = 0;

	ambient = rgbf(0.0f, 0.1f, 0.2f);
	diffuse = rgbf(0.1f, 0.5f, 0.7f);
	specular = rgbf(0.7f, 0.9f, 1.0f);
	shininess = 3.4f;

	Texture::setTexture(&diffuseTexture, "../textures/notex.png", NO_PURGE);
	Texture::setTexture(&noiseTexture, "../textures/notex.png", NO_PURGE);
}

/*---------------------------------------------------------------------------*/

MaterialWeird::~MaterialWeird()
{
	ensure_heap();

	if (diffuseTexture)
		sceneGraph.dereferenceResource(diffuseTexture);

	ensure_heap();

	if (noiseTexture)
		sceneGraph.dereferenceResource(noiseTexture);

	ensure_heap();
}

/*---------------------------------------------------------------------------*/

void MaterialWeird::prepareMaterial(
		FrameInfo_t			*di,
		Geometry			*geometry
	)
{
	u32 i;
	vec3f lpos;
	rgbf lcol;

	f32 time = di->deviceState->time;

	for (i = 0; i < MAX_SHADER_LIGHT_SLOTS; i++) {
		Light *light = geometry->getLight(i);

		if (light) {
			lpos.x = light->world.T0;
			lpos.y = light->world.T1;
			lpos.z = light->world.T2;
			lcol = light->getColor() * light->getIntensity();
			break;
		}
	}

	if (i == MAX_SHADER_LIGHT_SLOTS) {
		REPORT(WRN_NO_LIGHT);
		lpos = vec3f(0.0f, 0.0f, 0.0f);
		lcol = rgbf(1.0f, 1.0f, 1.0f);
	}

	shaderProgram->setUniformVector("LightColor", lcol.s, 1, 3);
	shaderProgram->setUniformVector("LightPosition", lpos.s, 1, 3);

	shaderProgram->setUniformVector("AmbientColor", ambient.s, 1, 3);
	shaderProgram->setUniformVector("DiffuseColor", diffuse.s, 1, 3);
	shaderProgram->setUniformVector("SpecularColor", specular.s, 1, 3);
	shaderProgram->setUniform("Shininess", &shininess, 1);

	time *= 0.2f;
	time = time - ((f32) ((i32) time));

	shaderProgram->setUniform("Time", &time, 1);

	if (diffuseTexture) {
		shaderProgram->setTextureSampler("DiffuseSampler", diffuseTexture, 0);
	} else {
		REPORT(WRN_NO_TEX);
	}

	if (noiseTexture) {
		shaderProgram->setTextureSampler("NoiseSampler", noiseTexture, 1);
	} else {
		REPORT(WRN_NO_TEX);
	}

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void MaterialWeird::setDiffuseTexture(
		Texture			*tex,

		bool			instantiate_resource
	)
{
	Texture::setTexture(&diffuseTexture, tex, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void MaterialWeird::setDiffuseTexture(
		const char		*filename,

		u32				purge_level
	)
{
	Texture::setTexture(&diffuseTexture, filename, purge_level);
}

/*---------------------------------------------------------------------------*/

Texture *MaterialWeird::getDiffuseTexture()
{
	return diffuseTexture;
}

/*---------------------------------------------------------------------------*/

void MaterialWeird::setNoiseTexture(
		Texture				*tex,

		bool				instantiate_resource
	)
{
	Texture::setTexture(&noiseTexture, tex, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void MaterialWeird::setNoiseTexture(
		const char			*filename,

		u32					purge_level
	)
{
	Texture::setTexture(&noiseTexture, filename, purge_level);
}

/*---------------------------------------------------------------------------*/

Texture *MaterialWeird::getNoiseTexture()
{
	return noiseTexture;
}


/*---------------------------------------------------------------------------*/

Resource *MaterialWeird::clone(
		const char		*clone_name,

		u32				purge_level
	)
{
	MaterialWeird *new_res = (MaterialWeird *) sceneGraph.createMaterial(materialType,
		clone_name, purge_level);

	new_res->ambient = ambient;
	new_res->diffuse = diffuse;
	new_res->shininess = shininess;
	new_res->specular = specular;

	new_res->setDiffuseTexture(diffuseTexture, true);
	new_res->setNoiseTexture(noiseTexture, true);

	return new_res;
}

/*---------------------------------------------------------------------------*/

