#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Node.h"

/* Warnings */
static const Result_t WRN_CAMERA = {
	RESULT_WARNING, "Camera resides in a non-active node" };

static const Result_t WRN_DETACHED = {
	RESULT_WARNING, "This node hasn't been detached from its parent" };

static const Result_t WRN_NOT_CONN = {
	RESULT_WARNING, "The two nodes are not connected" };

static const Result_t WRN_NO_PARENT = {
	RESULT_WARNING, "The node has no parent" };

/*---------------------------------------------------------------------------*/

Node::Node(
		const char		*nname,
		NodeType_t		ntype
	)
{
	nChildren = 0;
	parent = 0;
	firstChild = 0;
	next = 0;

	isActive = true;
	isVisible = true;

	type = ntype;

	name = duplicateString(nname);
}

/*---------------------------------------------------------------------------*/

Node::~Node()
{
	u32 dltd = 0;
	Node *c = firstChild;

	while (c) {
		Node *d = c->next;
		detachChild(c);
		delete c;
		dltd++;
		c = d;
	}

	free(name);
}

/*---------------------------------------------------------------------------*/

#define NODE_MAX_NAME_LENGTH			256

void Node::setName(
		char			*nname
	)
{
	u32 len = strlen(nname);

	free(name);

	if (len > NODE_MAX_NAME_LENGTH) {
		name = duplicateString(&nname[len] - NODE_MAX_NAME_LENGTH + 1);
	} else {
		name = duplicateString(nname);
	}
}

/*---------------------------------------------------------------------------*/

u32 Node::getChildrenCount()
{
	return nChildren;
}

/*---------------------------------------------------------------------------*/

void Node::attachChild(
		Node			*node
	)
{
	Node *c;

	if (node->parent) {
		REPORT(WRN_DETACHED);

		if (node->parent == this)
			return;

		node->parent->detachChild(node);
	}

	c = firstChild;
	firstChild = node;
	node->next = c;
	node->parent = this;

	nChildren++;
}

/*---------------------------------------------------------------------------*/

void Node::detachChild(
		Node			*node
	)
{
	Node *c = firstChild;
	Node *p = 0;

	while (c) {
		if (c == node) {
			c->parent = 0;

			if (p) {
				p->next = c->next;
			} else {
				firstChild = c->next;
			}
			break;
		}

		p = c;
		c = c->next;
	}

	if (!c) {
		REPORT(WRN_NOT_CONN);
	} else {
		c->next = 0;
		nChildren--;
	}
}

/*---------------------------------------------------------------------------*/

void Node::detachFromParent()
{
	if (parent)
		parent->detachChild(this);
}

/*---------------------------------------------------------------------------*/

Node *Node::getChild(
		u32			index
	)
{
	Node *node = firstChild;
	u32 i = 0;

	while (i < index && node) {
		node = node->next;
		i++;
	}

	return node;
}

/*---------------------------------------------------------------------------*/

Node *Node::getChild(
		char		*search_name,
		u32			max_depth
	)
{
	Node *node = firstChild;

	while (node) {

		if (strcmp(node->getName(), search_name) == 0)
			return node;

		node = node->next;
	}

	if (max_depth == 0)
		return 0;

	node = firstChild;

	while (node) {

		if (node->getChild(search_name, max_depth - 1) != 0)
			return node;

		node = node->next;
	}

	return 0;
}

/*---------------------------------------------------------------------------*/

Node *Node::getNextSibling()
{
	return next;
}

/*---------------------------------------------------------------------------*/

Node *Node::getParent()
{
	return parent;
}

/*---------------------------------------------------------------------------*/

void Node::update(
		FrameInfo_t			*di
	)
{
	if (isActive) {
		Node *c = firstChild;

		/* Update children */
		while (c) {
			c->update(di);

			c = c->next;
		}
	}
}

/*---------------------------------------------------------------------------*/

void Node::draw(
		FrameInfo_t			*di
	)
{
	if (type != NODE_WORLD && !parent) {
		REPORT(WRN_NO_PARENT);
	}

	if (isActive) {
		Node *c = firstChild;

		if (isVisible)
			drawSelf(di);

		/* Draw children */
		while (c) {
			c->draw(di);

			c = c->next;
		}
	}
}

/*---------------------------------------------------------------------------*/

void Node::transformCamera(
		mat4f		*matrix
	)
{
	if (!isActive)
		REPORT(WRN_CAMERA);

	parent->transformCamera(matrix);
}

/*---------------------------------------------------------------------------*/

void Node::setActive(
		bool		active
	)
{
	isActive = active;
}

/*---------------------------------------------------------------------------*/

void Node::setVisible(
		bool		visible
	)
{
	isVisible = visible;
}

/*---------------------------------------------------------------------------*/

char *Node::getName()
{
	return name;
}

/*---------------------------------------------------------------------------*/

void Node::dereferenceLight(
		Light				*light
	)
{
	Node *c = firstChild;

	while (c) {
		c->dereferenceLight(light);

		c = c->next;
	}
}

/*---------------------------------------------------------------------------*/

Node *Node::duplicate(
		char		*new_name,
		bool		recursively		/* True = recursively duplicate all
									 * children as well
									 */
	)
{
	Node *n = duplicateSelf();

	if (new_name) {
		n->setName(new_name);
	}

	if (recursively) {
		Node *c = firstChild;

		while (c) {
			n->attachChild(c->duplicate(0, true));
			c = c->next;
		}
	}

	return n;
}

/*---------------------------------------------------------------------------*/

Node *Node::duplicateSelf()
{
	return 0;
}

/*---------------------------------------------------------------------------*/

