#include "RenderChimp.h"
#include "SceneGraph.h"
#include "TriangleMesh.h"
#include "MaterialPoint.h"
#include "pwn.h"
#include "PwnParser.h"

u32 my_pwn_version = 1;

u16 load_counter = 0;

/* Warnings */
static const Result_t WRN_OPEN = {
	RESULT_WARNING, "Couldn't open .pwn file"};

static const Result_t WRN_FORMAT = {
	RESULT_WARNING, "File doesn't appear to be a .pwn file"};

static const Result_t WRN_VERSION = {
	RESULT_WARNING, "The .pwn version is not supported"};

/*---------------------------------------------------------------------------*/

static u8 read_u8(FILE *f)
{
	return (u8) fgetc(f);
}

/*---------------------------------------------------------------------------*/

static u16 read_u16(FILE *f)
{
	u16 r;

	r = ((u16) read_u8(f));
	r = r + (((u16) read_u8(f)) << 8);

	return r;
}

/*---------------------------------------------------------------------------*/

static u32 read_u32(FILE *f)
{
	u32 r;

	r = ((u32) read_u8(f));
	r += ((u32) read_u8(f)) << 8;
	r += ((u32) read_u8(f)) << 16;
	r += ((u32) read_u8(f)) << 24;

	return r;
}

/*---------------------------------------------------------------------------*/

static f32 read_f32(FILE *f)
{
	f32 v;
	u8 r[4];

	r[0] = read_u8(f);
	r[1] = read_u8(f);
	r[2] = read_u8(f);
	r[3] = read_u8(f);

	v = *((f32 *) r);

	return v;
}

/*---------------------------------------------------------------------------*/

static void read_str(FILE *f, char *target, i32 n)
{
	i32 i;

	for (i = 0; i < n; i++) {
		i32 c = fgetc(f);
		target[i] = (char) c;
	}
}

/*---------------------------------------------------------------------------*/

/* TODO: Support more types of materials */
static Material *parse_material(
		FILE		*f,
		bool		reload_resources,
		u32			purge_level
	)
{
	u32 i;
	char buf[1024];
	char id_buf[PWN_ID_MAX_LENGTH];

	vec3f ambient = vec3f(0, 0, 0);
	vec3f diffuse = vec3f(0, 0, 0);
	vec3f specular = vec3f(0, 0, 0);
	float shininess = 0;
	float emissive = 0;
	float specular_level = 0;

	MaterialPoint *mat;

	for (;;) {
		read_str(f, buf, 4);

		if (strncmp(buf, PWN_MATERIAL_ID, 4) == 0) {
			u32 len = read_u32(f);
			for (i = 0; i < len && i < 1000; i++) {
				id_buf[i] = (char) read_u8(f);
			}

			if (reload_resources) {
				sprintf(&id_buf[i], "%x", load_counter);
				i += 4;
			}
			id_buf[i] = '\0';
		} else if (strncmp(buf, PWN_MATERIAL_AMB_COLOR_PREFIX, 4) == 0) {
			ambient.x = read_f32(f);
			ambient.y = read_f32(f);
			ambient.z = read_f32(f);
		} else if (strncmp(buf, PWN_MATERIAL_DIF_COLOR_PREFIX, 4) == 0) {
			diffuse.x = read_f32(f);
			diffuse.y = read_f32(f);
			diffuse.z = read_f32(f);
		} else if (strncmp(buf, PWN_MATERIAL_SPC_COLOR_PREFIX, 4) == 0) {
			specular.x = read_f32(f);
			specular.y = read_f32(f);
			specular.z = read_f32(f);
		} else if (strncmp(buf, PWN_MATERIAL_SHI_COLOR_PREFIX, 4) == 0) {
			shininess = read_f32(f);
		} else if (strncmp(buf, PWN_MATERIAL_EMI_COLOR_PREFIX, 4) == 0) {
			emissive = read_f32(f);
		} else if (strncmp(buf, PWN_MATERIAL_SPL_COLOR_PREFIX, 4) == 0) {
			specular_level = read_f32(f);
		} else {
			break;
		}
	}

	fseek(f, -4, SEEK_CUR); /* Rewind last string tag read */

	mat = (MaterialPoint *) sceneGraph.createMaterial(
		MATERIAL_POINT, id_buf, purge_level);

	mat->ambient = ambient;
	mat->diffuse = diffuse;
	mat->specular = specular;
	mat->shininess = shininess;
	mat->emissive = emissive;
	mat->specularLevel = specular_level;

	return mat;
}

/*---------------------------------------------------------------------------*/

static TriangleMesh *parse_mesh(
		FILE				*f,
		u32					purge_level,
		Material			**material_list,
		u32					n_materials,
		bool				reload_resources
	)
{
	i32 material_index = -1;
	u32 i, buflen;
	char buf[1024];
	char id_buf[PWN_ID_MAX_LENGTH];

	u32 atr_offset[PWN_MESH_MAX_ATTRIBUTES];
	u32 atr_length[PWN_MESH_MAX_ATTRIBUTES];
	char *atr_name[PWN_MESH_MAX_ATTRIBUTES];
	u32 n_atr = 0;

	u32 n_indices = 0;
	u32 n_vertices = 0;
	u32 stride = 0;

	TriangleMesh *mesh;
	VertexArray *va;
	IndexArray *ia;

	Material *mat = 0;

	u16 *index_array = 0;
	f32 *vertex_array = 0;

	for (;;) {
		read_str(f, buf, 4);

		if (strncmp(buf, PWN_MESH_ID, 4) == 0) {
			buflen = read_u32(f);

			for (i = 0; i < buflen && i < 1000; i++) {
				id_buf[i] = (char) read_u8(f);
			}

			if (reload_resources) {
				sprintf(&id_buf[i], "%x", load_counter);
				i += 4;
			}

			id_buf[i] = '\0';

		} else if (strncmp(buf, PWN_MESH_MATERIAL, 4) == 0) {

			material_index = (i32) read_u32(f);

		} else if (strncmp(buf, PWN_MESH_ATTRIBUTES, 4) == 0) {
			u32 j;
			n_atr = read_u32(f);

			for (i = 0; i < n_atr; i++) {
				u32 len = read_u32(f);

				atr_name[i] = (char *) malloc(sizeof(char) * (len + 1));

				for (j = 0; j < len; j++) {
					atr_name[i][j] = (char) read_u8(f);
					if (atr_name[i][j] >= 'A' && atr_name[i][j] <= 'Z' && j != 0) {
						atr_name[i][j] -= ('A' - 'a');
					}
				}
				atr_name[i][j] = '\0';

				atr_offset[i] = read_u32(f);
				atr_length[i] = read_u32(f);
			}

		} else if (strncmp(buf, PWN_MESH_INDEX_ARRAY, 4) == 0) {
			u32 type_size;

			n_indices = read_u32(f);

			index_array = (u16 *) malloc(sizeof(u16) * n_indices);
			
			type_size = read_u32(f);

			if (type_size == 4) {
				for (i = 0; i < n_indices; i++) {
					index_array[i] = (u16) read_u32(f);
					/* TODO: This should generate an error! */
				}
			} else if (type_size == 2) {
				for (i = 0; i < n_indices; i++) {
					index_array[i] = read_u16(f);
				}
			}
		} else if (strncmp(buf, PWN_MESH_VERTEX_ARRAY, 4) == 0) {
			n_vertices = read_u32(f);
			stride = read_u32(f);

			vertex_array = (f32 *) malloc(sizeof(f32) * n_vertices * stride);

			for (i = 0; i < n_vertices * stride; i++) {
				vertex_array[i] = read_f32(f);
			}
		} else {
			break;
		}
	}

	fseek(f, -4, SEEK_CUR); /* Rewind last string tag read */

	ia = sceneGraph.createIndexArray(id_buf, index_array, n_indices,
		TRIANGLES, purge_level);

	free(index_array);

	va = sceneGraph.createVertexArray(id_buf, vertex_array, stride,
		n_vertices, purge_level);

	free(vertex_array);

	for (i = 0; i < n_atr; i++)
		va->setAttribute(atr_offset[i], atr_length[i], atr_name[i]);

	if (material_index > -1)
		mat = material_list[material_index];

	for (i = 0; i < va->getNAttributes(); i++) {
		if (strcmp(va->getAttributeName(i), "Normal") == 0) {
			break;
		}
	}

	if (i == va->getNAttributes())
		va->generateNormals(ia);

	for (i = 0; i < n_atr; i++)
		free(atr_name[i]);

	mesh = sceneGraph.createTriangleMesh(id_buf, va, ia, mat, false);

	return mesh;
}

/*---------------------------------------------------------------------------*/

static void parse_child(
		FILE				*f,
		Transformable		*parent,

		TriangleMesh		**mesh_list,
		u32					n_meshes
	)
{
	char buf[4];
	u32 index;

	vec3f rotate;
	f32 rotate_angle = 0.0f;
	vec3f scale;
	vec3f translate;

	bool has_rotate = false;
	bool has_scale = false;
	bool has_translate = false;

	for (;;) {
		read_str(f, buf, 4);

		if (strncmp(buf, PWN_NODE_PREFIX, 4) == 0) {
			
		} else if (strncmp(buf, PWN_NODE_ROTATE, 4) == 0) {
			has_rotate = true;

			rotate.x = read_f32(f);
			rotate.y = read_f32(f);
			rotate.z = read_f32(f);
			rotate_angle = read_f32(f);

		} else if (strncmp(buf, PWN_NODE_TRANSLATE, 4) == 0) {
			has_translate = true;

			translate.x = read_f32(f);
			translate.y = read_f32(f);
			translate.z = read_f32(f);

		} else if (strncmp(buf, PWN_NODE_SCALE, 4) == 0) {
			has_scale = true;

			scale.x = read_f32(f);
			scale.y = read_f32(f);
			scale.z = read_f32(f);

		} else if (strncmp(buf, PWN_NODE_TYPE_MESH, 4) == 0) {
			u32 i;
			Transformable *node;
			u32 n_children;

			index = read_u32(f);
			n_children = read_u32(f);

			node = mesh_list[index];

			if (has_translate)
				node->translate(translate);

			if (has_rotate)
				node->rotate(rotate_angle, rotate);

			if (has_scale)
				node->scale(scale);

			for (i = 0; i < n_children; i++) {
				parse_child(f, node, mesh_list, n_meshes);
			}

			parent->attachChild(node);

			break;

		} else {
			break;
		}
	}
}

/*---------------------------------------------------------------------------*/

static Group *parse_node_tree(
		FILE				*f,
		Group				*root_group,

		TriangleMesh		**mesh_list,
		u32					n_meshes
	)
{
	parse_child(f, root_group, mesh_list, n_meshes);

	fseek(f, -4, SEEK_CUR); /* Rewind last string tag read */

	return root_group;
}

/*---------------------------------------------------------------------------*/

void PwnParse(
		Group				*target,
		const char			*filename,
		bool				first_mesh_only,
		bool				reload_resources,
		u32					purge_level
	)
{
	u32 v;
	u32 n_meshes = 0, n_materials = 0;
	u32 total_meshes, total_materials;
	char buf[1024];

	TriangleMesh *mesh_list[PWN_MESH_MAX_MESHES];
	Material *material_list[PWN_MESH_MAX_MATERIALS];

	FILE *f = fopen(filename, "rb");

	if (!f) {
		REPORT(WRN_OPEN);
		return;
	}

	read_str(f, buf, 4);

	if (strncmp(buf, PWN_ID, 4) != 0) {
		REPORT(WRN_FORMAT);
		return;
	}

	v = read_u32(f);

	if (my_pwn_version != PWN_VERSION) {
		REPORT(WRN_VERSION);
		return;
	}

	total_materials = read_u32(f);
	total_meshes = read_u32(f);

	for (;;) {
		read_str(f, buf, 4);

		if (strncmp(buf, PWN_MATERIAL_PREFIX, 4) == 0) {

			material_list[n_materials] = parse_material(f, reload_resources, purge_level);
			n_materials++;

		} else if (strncmp(buf, PWN_MESH_PREFIX, 4) == 0) {

			mesh_list[n_meshes] = parse_mesh(f, purge_level, material_list, n_materials, reload_resources);
			n_meshes++;

			if (first_mesh_only) {
				target->attachChild(mesh_list[0]);
				break;
			}

		} else if (strncmp(buf, PWN_NODE_PREFIX, 4) == 0) {

			parse_node_tree(f, target, mesh_list, n_meshes);

		} else if (strncmp(buf, PWN_END_FILE, 4) == 0) {

			break;

		}
	}

	load_counter++;

	fclose(f);
}

/*---------------------------------------------------------------------------*/

