#include "RenderChimp.h"
#include "SceneGraph.h"
#include "RenderState.h"


/* States that can be enabled or disabled */
#define STATE_DEPTH					0x01
#define STATE_STENCIL				0x02
#define STATE_BLEND					0x04
#define STATE_CULL					0x08
#define STATE_SCISSOR				0x10

const u32 gl_src_blend_map[8] = {
	GL_ZERO,
	GL_ONE,
	GL_SRC_COLOR,
	GL_ONE_MINUS_SRC_COLOR,
	GL_SRC_ALPHA,
	GL_ONE_MINUS_SRC_ALPHA,
	GL_DST_ALPHA,
	GL_ONE_MINUS_DST_ALPHA
};

const u32 gl_dst_blend_map[8] = {
	GL_ZERO,
	GL_ONE,
	GL_DST_COLOR,
	GL_ONE_MINUS_DST_COLOR,
	GL_SRC_ALPHA_SATURATE,
	GL_ONE_MINUS_SRC_ALPHA,
	GL_DST_ALPHA,
	GL_ONE_MINUS_DST_ALPHA
};

const u32 gl_cmp_map[8] = {
	GL_NEVER,
	GL_LESS,
	GL_EQUAL,
	GL_LEQUAL,
	GL_GREATER,
	GL_NOTEQUAL,
	GL_GEQUAL,
	GL_ALWAYS
};

const u32 gl_cull_map[3] = {
	GL_FRONT,
	GL_BACK,
	GL_FRONT_AND_BACK
};

/*---------------------------------------------------------------------------*/

/* Global current render state */
static StateSettings_t current_settings;

/*---------------------------------------------------------------------------*/

RenderState::RenderState()
{
	settings.changes = 0;
	settings.flags = 0;
}

/*---------------------------------------------------------------------------*/

RenderState::~RenderState()
{

}

/*---------------------------------------------------------------------------*/

inline i32 toggle(
		u8				changes,
		u8				flags,
		u32				old_flags,
		u32				mask,
		u32				gl_def,
		bool			force_set
	)
{
	if (!(changes & mask))
		return 0;

	if ((flags & mask) != (old_flags & mask) || force_set) {
		current_settings.flags &= 0xFF ^ mask;
		current_settings.flags |= (flags & mask);

		if (flags & mask) {
			glEnable(gl_def);
			GL_ASSERT();
			return 1;
		} else {
			glDisable(gl_def);
			GL_ASSERT();
			return 0;
		}
	}

	return (flags & mask);
}

/*---------------------------------------------------------------------------*/

void RenderState::setState(
		StateSettings_t			*curr,
		u8						changes,
		bool					force_set
	)
{
	StateSettings_t *old = &current_settings;

	if (toggle(changes, curr->flags, old->flags, STATE_DEPTH, GL_DEPTH_TEST, force_set)) {
		glDepthFunc(gl_cmp_map[(u32) curr->depthTest]);
		current_settings.depthTest = curr->depthTest;
	}

	GL_ASSERT();

	if (toggle(changes, curr->flags, old->flags, STATE_STENCIL, GL_STENCIL_TEST, force_set)) {
		glStencilFunc(gl_cmp_map[(u32) curr->stencilTest], curr->stencilRef, curr->stencilMask);
		current_settings.stencilRef = curr->stencilRef;
		current_settings.stencilMask = curr->stencilMask;
	}

	GL_ASSERT();

	if (toggle(changes, curr->flags, old->flags, STATE_BLEND, GL_BLEND, force_set)) {
		glBlendFunc(gl_src_blend_map[(u32) curr->srcBlend], gl_dst_blend_map[(u32) curr->dstBlend]);
		current_settings.srcBlend = curr->srcBlend;
		current_settings.dstBlend = curr->dstBlend;
	}

	GL_ASSERT();

	if (toggle(changes, curr->flags, old->flags, STATE_CULL, GL_CULL_FACE, force_set)) {
		glCullFace(gl_cull_map[(u32) curr->cull]);
		current_settings.cull = curr->cull;
	}

	GL_ASSERT();

	if (toggle(changes, curr->flags, old->flags, STATE_SCISSOR, GL_SCISSOR_TEST, force_set)) {
		glScissor(curr->scissorX, curr->scissorY, curr->scissorW, curr->scissorH);
		current_settings.scissorX = curr->scissorX;
		current_settings.scissorY = curr->scissorY;
		current_settings.scissorW = curr->scissorW;
		current_settings.scissorH = curr->scissorH;
	}

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void RenderState::enableDepth(
		StateCompare_t	depth_test
	)
{
	settings.changes |= STATE_DEPTH;
	settings.flags |= STATE_DEPTH;

	settings.depthTest = depth_test;
}

/*---------------------------------------------------------------------------*/

void RenderState::disableDepth()
{
	settings.changes |= STATE_DEPTH;
	settings.flags &= ~STATE_DEPTH;
}

/*---------------------------------------------------------------------------*/

void RenderState::enableStencil(
		StateCompare_t	stencil_test,
		i32				ref,
		u32				mask
	)
{
	settings.changes |= STATE_STENCIL;
	settings.flags |= STATE_STENCIL;

	settings.stencilRef = ref;
	settings.stencilMask = mask;
}

/*---------------------------------------------------------------------------*/

void RenderState::disableStencil()
{
	settings.changes |= STATE_STENCIL;
	settings.flags &= ~STATE_STENCIL;
}

/*---------------------------------------------------------------------------*/

void RenderState::enableBlend(
		StateSrcBlend_t	src_blend,
		StateDstBlend_t	dst_blend
	)
{
	settings.changes |= STATE_BLEND;
	settings.flags |= STATE_BLEND;

	settings.srcBlend = src_blend;
	settings.dstBlend = dst_blend;
}

/*---------------------------------------------------------------------------*/

void RenderState::disableBlend()
{
	settings.changes |= STATE_BLEND;
	settings.flags &= ~STATE_BLEND;
}

/*---------------------------------------------------------------------------*/

void RenderState::enableCulling(
		StateCulling_t	cull_strategy
	)
{
	settings.changes |= STATE_CULL;
	settings.flags |= STATE_CULL;
	
	settings.cull = cull_strategy;
}

/*---------------------------------------------------------------------------*/

void RenderState::disableCulling()
{
	settings.changes |= STATE_CULL;
	settings.flags &= ~STATE_CULL;
}

/*---------------------------------------------------------------------------*/

void RenderState::enableScissor(
		i32				x,
		i32				y,
		u32				width,
		u32				height
	)
{
	settings.changes |= STATE_SCISSOR;
	settings.flags |= STATE_SCISSOR;
	
	settings.scissorX = x;
	settings.scissorY = y;
	settings.scissorW = width;
	settings.scissorH = height;
}

/*---------------------------------------------------------------------------*/

void RenderState::disableScissor()
{
	settings.changes |= STATE_SCISSOR;
	settings.flags &= ~STATE_SCISSOR;
}

/*---------------------------------------------------------------------------*/

void RenderState::clear()
{
	settings.flags = 0;
}

/*---------------------------------------------------------------------------*/

StateSettings_t *RenderState::getCurrentState()
{
	return &current_settings;
}

/*---------------------------------------------------------------------------*/

