#include "RenderChimp.h"
#include "SceneGraph.h"


/*---------------------------------------------------------------------------*/

SceneGraph::SceneGraph()
{
	textIsAlive = false;
}

/*---------------------------------------------------------------------------*/

SceneGraph::~SceneGraph()
{
}

/*---------------------------------------------------------------------------*/

void SceneGraph::init()
{
	materialList = new ResourceList;
	shaderProgramList = new ResourceList;

	vertexList = new ResourceList;
	indexList = new ResourceList;
	textureList = new ResourceList;
	cubeMapList = new ResourceList;
	shaderList = new ResourceList;

	text = new Text("../fonts/font.png", 4);

	textIsAlive = true;
}

/*---------------------------------------------------------------------------*/

void SceneGraph::destroy()
{
	delete materialList;
	delete shaderProgramList;

	delete vertexList;
	delete indexList;
	delete textureList;
	delete cubeMapList;
	delete shaderList;

	delete text;
}

/*---------------------------------------------------------------------------*/

/* Common destroyer */
void SceneGraph::deleteNode(
		Node			*node
	)
{
	delete node;
}

/*---------------------------------------------------------------------------*/

/* World */
World *SceneGraph::createWorld(
		const char			*nname
	)
{
	return new World(nname);
}

/*---------------------------------------------------------------------------*/

/* State */
State *SceneGraph::createState(
		const char			*nname
	)
{
	return new State(nname);
}

/*---------------------------------------------------------------------------*/

/* Transformable */

/* Group */
Group *SceneGraph::createGroup(
		const char			*nname
	)
{
	return new Group(nname);
}

/*---------------------------------------------------------------------------*/

Group *SceneGraph::createGroup(
		const char			*nname,
		const char			*filename,
		bool			reload_resources,
		u32				purge_level
	)
{
	return new Group(nname, filename, reload_resources, purge_level);
}

/*---------------------------------------------------------------------------*/

/* Camera */
Camera *SceneGraph::createCamera(
		const char			*nname
	)
{
	return new Camera(nname);
}

/*---------------------------------------------------------------------------*/

/* Light */
Light *SceneGraph::createLight(
		const char			*nname
	)
{
	return new Light(nname);
}

/*---------------------------------------------------------------------------*/

Light *SceneGraph::createLight(
		const char			*nname,
		rgbf			color,
		f32				intensity
	)
{
	return new Light(nname, color, intensity);
}

/*---------------------------------------------------------------------------*/

TriangleMesh *SceneGraph::createTriangleMesh(
		const char			*nname,
		const char			*filename,

		bool			reload_resources,
		u32				purge_level
	)
{
	return new TriangleMesh(
		nname,
		filename,
		reload_resources,
		purge_level
	);
}

/*---------------------------------------------------------------------------*/

TriangleMesh *SceneGraph::createTriangleMesh(
		const char			*nname,

		VertexArray		*vertex_array,
		IndexArray		*index_array,
		Material		*material,

		bool			instantiate_resources
	)
{
	return new TriangleMesh(
		nname,

		vertex_array,
		index_array,
		material,

		instantiate_resources
	);
}

/*---------------------------------------------------------------------------*/

TriangleMesh *SceneGraph::createTriangleMesh(
		const char					*nname,

		vec3f					*position_array,
		vec2f					*texcoord_array,
		vec3f					*normal_array,
		vec4f					*color_array,
		u32						n_vertices,

		u16						*index_array,
		u32						n_indices,
		IndexArrayLayout_t		index_layout,

		u32						purge_level
	)
{
	return new TriangleMesh(
		nname,

		position_array,
		texcoord_array,
		normal_array,
		color_array,
		n_vertices,

		index_array,
		n_indices,
		index_layout,

		purge_level
	);
}

/*---------------------------------------------------------------------------*/

/* Sprite */

Sprite *SceneGraph::createSprite(
		const char				*nname,
		Texture					*ntexture,

		bool					instantiate_resources
	)
{
	return new Sprite(nname, ntexture, instantiate_resources);
}

/*---------------------------------------------------------------------------*/

Sprite *SceneGraph::createSprite(
		const char				*nname,
		const char				*texture_filename,

		u32						purge_level
	)
{
	return new Sprite(nname, texture_filename, purge_level);
}

/*---------------------------------------------------------------------------*/

bool SceneGraph::dereferenceResource(
		Resource		*resource
	)
{

	GL_ASSERT();

	if (resource->purgeLevel == 0) {
		ResourceType_t type = resource->getType();

		switch (type) {
			case RESOURCE_VERTEX_ARRAY:
				return vertexList->remove(resource);
				break;
			case RESOURCE_INDEX_ARRAY:
				return indexList->remove(resource);
				break;
			case RESOURCE_TEXTURE:
				return textureList->remove(resource);
				break;
			case RESOURCE_CUBEMAP:
				return cubeMapList->remove(resource);
				break;
			case RESOURCE_SHADER:
				return shaderList->remove(resource);
				break;
			case RESOURCE_SHADER_PROGRAM:
				return shaderProgramList->remove(resource);
				break;
			case RESOURCE_MATERIAL:
				return materialList->remove(resource);
				break;
			default:
				break;
		}
	} else {
		if (resource->n_refs > 0) {
			resource->n_refs--;
		}
	}


	return false;
}

/*---------------------------------------------------------------------------*/

void SceneGraph::purgeResources(
		u32				max_purge_level
	)
{
	/* These have dependencies to other resources = delete first */
	materialList->purge(max_purge_level);
	shaderProgramList->purge(max_purge_level);

	/* These do not have any dependencies */
	vertexList->purge(max_purge_level);
	indexList->purge(max_purge_level);
	textureList->purge(max_purge_level);
	cubeMapList->purge(max_purge_level);
	shaderList->purge(max_purge_level);

}

/*---------------------------------------------------------------------------*/

VertexArray *SceneGraph::createVertexArray(
		const char				*nidentifier,

		f32						*new_array,
		u32						nstride,
		u32						n,
		u32						purge_level
	)
{
	VertexArray *a = (VertexArray *) vertexList->fetch(nidentifier);

	if (!a) {
		a = new VertexArray(nidentifier, new_array, nstride,
			n, purge_level);

		vertexList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

VertexArray *SceneGraph::getVertexArray(
		const char					*identifier
	)
{
	return (VertexArray *) vertexList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

IndexArray *SceneGraph::createIndexArray(
		const char				*nidentifier,

		u16						*new_array,
		u32						n,
		IndexArrayLayout_t		layout,
		u32						purge_level
	)
{
	IndexArray *a = (IndexArray *) indexList->fetch(nidentifier);

	if (!a) {
		a = new IndexArray(nidentifier, new_array, n, layout, purge_level);

		indexList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

IndexArray *SceneGraph::getIndexArray(
		const char				*identifier
	)
{
	return (IndexArray *) indexList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

Texture *SceneGraph::createTexture(
		const char				*filename,
		bool					generate_mipmaps,
		TextureFilter_t			filter,
		TextureWrap_t			wrap,
		u32						purge_level
	)
{
	Texture *a = (Texture *) textureList->fetch(filename);

	if (!a) {
		a = new Texture(filename, generate_mipmaps, filter, wrap, purge_level);
		textureList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

Texture *SceneGraph::createTexture(
		u8						*pixels,
		const char				*nidentifier,
		TextureFormat_t			nformat,
		u32						nwidth,
		u32						nheight,
		bool					generate_mipmaps,
		TextureFilter_t			filter,
		TextureWrap_t			wrap,
		u32						purge_level
	)
{
	Texture *a = (Texture *) textureList->fetch(nidentifier);

	if (!a) {
		a = new Texture(pixels, nidentifier, nformat, nwidth, nheight,
			generate_mipmaps, filter, wrap, purge_level);

		textureList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

CubeMap *SceneGraph::createCubeMap(
		const char			*nidentifier,

		const char			*nleft,
		const char			*nright,
		const char			*nup,
		const char			*ndown,
		const char			*nfront,
		const char			*nback,

		u32				purge_level
	)
{
	CubeMap *a = (CubeMap *) cubeMapList->fetch(nidentifier);

	if (!a) {
		a = new CubeMap(nidentifier,
			nleft, nright, nup, ndown, nfront, nback,
			purge_level);

		cubeMapList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

Texture *SceneGraph::getTexture(
		const char				*identifier
	)
{
	return (Texture *) textureList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

CubeMap *SceneGraph::getCubeMap(
		const char				*identifier
	)
{
	return (CubeMap *) cubeMapList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

Material *SceneGraph::createMaterial(
		MaterialType_t			material_type,
		const char				*nidentifier,
		u32						purge_level
	)
{
	Material *a = (Material *) materialList->fetch(nidentifier);

	if (!a) {
		a = Material::create(material_type, nidentifier, purge_level);

		materialList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

Material *SceneGraph::getMaterial(
		const char				*identifier
	)
{
	return (Material *) materialList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

ShaderProgram *SceneGraph::createShaderProgram(
		const char				*nidentifier,
		char					*shader_prefix,
		const char				*vertex_shader_filename,
		const char				*fragment_shader_filename,

		u32						purge_level
	)
{
	ShaderProgram *a = (ShaderProgram *) shaderProgramList->fetch(nidentifier);

	if (!a) {
		u32 lenf = strlen(vertex_shader_filename);
		u32 lenp = strlen(shader_prefix);
		char *new_key = (char *) malloc(lenf + lenp + 1);
		sprintf(new_key, "%s%s", shader_prefix, vertex_shader_filename);
		
		Shader *vs, *fs;

		vs = (Shader *) shaderList->fetch(new_key);

		GL_ASSERT();

		if (!vs) {
			vs = new Shader(new_key, shader_prefix, vertex_shader_filename, SHADER_TYPE_VERTEX, purge_level);
			shaderList->add(vs);
		} else {
			vs->setMaxPurgeLevel(purge_level);
		}
		
		free(new_key);
		
		lenf = strlen(fragment_shader_filename);
		new_key = (char *) malloc(lenf + lenp + 1);
		sprintf(new_key, "%s%s", shader_prefix, fragment_shader_filename);

		fs = (Shader *) shaderList->fetch(new_key);

		GL_ASSERT();

		if (!fs) {
			fs = new Shader(new_key, shader_prefix, fragment_shader_filename, SHADER_TYPE_FRAGMENT, purge_level);
			shaderList->add(fs);
		} else {
			fs->setMaxPurgeLevel(purge_level);
		}
		
		free(new_key);

		a = new ShaderProgram(nidentifier, vs, fs, purge_level);

		GL_ASSERT();

		shaderProgramList->add(a);
	} else {
		a->setMaxPurgeLevel(purge_level);
	}

	return a;
}

/*---------------------------------------------------------------------------*/

ShaderProgram *SceneGraph::getShaderProgram(
		const char				*identifier
	)
{
	return (ShaderProgram *) shaderProgramList->fetch(identifier);
}

/*---------------------------------------------------------------------------*/

/* HACK */
void SceneGraph::resize(
		i32 x,
		i32 y
	)
{
	text->resize((f32) x, (f32) y);
}

/*---------------------------------------------------------------------------*/

#ifdef DEBUG

void SceneGraph::printDebugInfo()
{
	ensure_heap();

	materialList->printContents("Material");
	shaderProgramList->printContents("ShaderProgram");

	vertexList->printContents("VertexArray");
	indexList->printContents("IndexArray");
	textureList->printContents("Texture");
	cubeMapList->printContents("CubeMap");
	shaderList->printContents("Shader");

	ensure_heap();
}

#else

void SceneGraph::printDebugInfo()
{
}

#endif /* DEBUG */

/*---------------------------------------------------------------------------*/

SceneGraph sceneGraph;


