/*
 * This defines the Scene Graph structure.
 *
 * Use the create- and delete-functions from this function to handle all node
 * and resource allocations and deallocations.
 */

#ifndef __RC_SCENEGRAPH_H__
#define __RC_SCENEGRAPH_H__

#include "RenderChimp.h"
#include "ape.h"
#include "SceneGraphAux.h"

/*---------------------------------------------------------------------------*/

class SceneGraph {
	friend class TriangleMesh;

/*----------------------------------------------------------[ Functions ]----*/

	public:
		SceneGraph();
		~SceneGraph();

		void init();
		void destroy();

/*==============================================================[ Nodes ]====*/

/*---------------------------------------------------------[ Destroyers ]----*/

		/*
		 * Common destroyer
		 * Use this function when destroying a node, no matter what type it
		 * has. The children of this node will also be deleted recursively, so
		 * make sure that the node doesn't have any children that shouldn't be
		 * deleted.
		 */
		void deleteNode(
				Node			*node
			);

/*-----------------------------------------------------------[ Creators ]----*/

		/* World */
		World *createWorld(
				const char			*nname
			);


		/* State */
		State *createState(
				const char			*nname
			);


			/* Transformable */

		/* Group */
		Group *createGroup(
				const char			*nname
			);

		Group *createGroup(
				const char			*nname,
				const char			*filename,	/* .pwn file */
				bool			reload_resources, /* true = all resources are
												   * loaded with a new name,
												   * even if there are matching
												   * resources in the resource
												   * lists. false = uses
												   * existing resources, if
												   * available.
												   */
				u32				purge_level
			);


		/* Camera */
		Camera *createCamera(
				const char			*nname
			);


		/* Light */
		Light *createLight(
				const char			*nname
			);

		Light *createLight(
				const char			*nname,
				rgbf			color,
				f32				intensity
			);


			/* Geometry */

		/* TriangleMesh */

		/* Loads the first mesh found from the specified .pwn file */
		TriangleMesh *createTriangleMesh(
				const char			*nname,
				const char			*filename, /* /.pwn file */

				bool			reload_resources,
				u32				purge_level
			);

		TriangleMesh *createTriangleMesh(
				const char			*nname,

				VertexArray		*vertex_array,
				IndexArray		*index_array,
				Material		*material,

				bool			instantiate_resources
			);

		TriangleMesh *createTriangleMesh(
				const char			*nname,

				/* Any of these may be 0 */
				vec3f			*position_array,
				vec2f			*texcoord_array,
				vec3f			*normal_array,
				vec4f			*color_array,
				u32				n_vertices,

				u16				*index_array,
				u32				n_indices,
				IndexArrayLayout_t index_layout,

				u32				purge_level
			);


		/* Sprite */

		Sprite *createSprite(
				const char		*nname,
				Texture			*ntexture,

				bool			instantiate_resources
			);

		Sprite *createSprite(
				const char		*nname,
				const char		*texture_filename,

				u32				purge_level
			);


/*==========================================================[ Resources ]====*/


/*---------------------------------------------------------[ Destroyers ]----*/


		/*
		 * Common dereferencer.
		 *
		 * Use this to dereference and potentially delete resources of any type.
		 * When a resource is dereferenced its reference counter is decreased.
		 * When there are no more references to a resource, it's actually
		 * deallocated from memory.
		 * This only works for resources with purgeLevel = 0
		 * Returns true if the resource was actually deleted from memory.
		 */
		bool dereferenceResource(
				Resource		*resource
			);

		/*
		 * Destroy all resources of purgeLevel > 0 and
		 * purgeLevel <= max_purge_level.
		 */
		void purgeResources(
				u32				max_purge_level
			);

/*-----------------------------------------------------------[ Creators ]----*/

		/* VertexArray */

		VertexArray *createVertexArray(
				const char		*nidentifier,

				f32				*new_array,		/* Array of structs */
				u32				nstride,		/* Length of each struct */
				u32				n,				/* Number of structs in
												 * array
												 */
				u32				purge_level
			);


		/* IndexArray */

		IndexArray *createIndexArray(
				const char		*nidentifier,

				u16				*new_array,		/* Array of indices */
				u32				n,				/* Number of indices */

				IndexArrayLayout_t layout,		/* Triangles, triangle
												 * fan or triangle strip
												 */
				u32				purge_level
			);


		/* Texture */

		Texture *createTexture(
				const char		*filename,
				bool			generate_mipmaps,
				TextureFilter_t	filter,
				TextureWrap_t	wrap,
				u32				purge_level
			);

		Texture *createTexture(
				u8				*pixels,
				const char		*nidentifier,
				TextureFormat_t	nformat,
				u32				nwidth,				/* width in pixels */
				u32				nheight,			/* height in pixels */
				bool			generate_mipmaps,
				TextureFilter_t	filter,
				TextureWrap_t	wrap,
				u32				purge_level
			);

		/* CubeMap */

		CubeMap *createCubeMap(
				const char			*nidentifier,

				const char			*nleft,
				const char			*nright,
				const char			*nup,
				const char			*ndown,
				const char			*nfront,
				const char			*nback,

				u32				purge_level
			);


		/* Material */

		Material *createMaterial(
				MaterialType_t	material_type,
				const char		*nidentifier,
				u32				purge_level
			);


		/* ShaderProgram */

		ShaderProgram *createShaderProgram(
				const char		*nidentifier,
				char			*shader_prefix,
				const char		*vertex_shader_filename,
				const char		*fragment_shader_filename,

				u32				purge_level
			);


/*------------------------------------------------------------[ Getters ]----*/

		VertexArray *getVertexArray(
				const char		*identifier
			);

		IndexArray *getIndexArray(
				const char		*identifier
			);

		Texture *getTexture(
				const char		*identifier
			);

		CubeMap *getCubeMap(
				const char		*identifier
			);

		Material *getMaterial(
				const char		*identifier
			);

		ShaderProgram *getShaderProgram(
				const char		*identifier
			);


/*===========================================================================*/

		/* HACK */
		void resize(
				i32 x,
				i32 y
			);

		void printDebugInfo();


/*------------------------------------------------------------[ Members ]----*/

	public:
		Text				*text;
		bool				textIsAlive;

	private:
		ResourceList		*vertexList;
		ResourceList		*indexList;
		ResourceList		*textureList;
		ResourceList		*cubeMapList;
		ResourceList		*shaderList;
		ResourceList		*shaderProgramList;
		ResourceList		*materialList;

};

/* Singleton scene graph object */
extern SceneGraph sceneGraph;

#endif /* __RC_SCENEGRAPH_H__ */

