/*
 * Vector and Matrix operations (along with various other constants and maths
 * stuff...
 */

#ifndef __RC_VECTORMATH_H__
#define __RC_VECTORMATH_H__

/*==========================================================[ Constants ]====*/

#ifndef PI
	#define fPI					3.1415926535f
	#define f2PI				(2.0f * 3.1415926535f)
	#define fPIinv				(1.0f / 3.1415926535f)
#endif

/*=============================================================[ Scalar ]====*/

#define sclamp(s, a, b)			((s) > (b) ? (b) : ((s) < (a) ? (a) : (s)))

#define ssaturate(s)			sclamp((s), 0.0f, 1.0f)

#define sswap(a, b)				T tmp##a##b = (a); (a) = (b); (b) = tmp##a##b

#define toRad(a)				((a / 360.0f) * f2PI)

#define toDeg(a)				((a * 360.0f) / f2PI)

/*===========================================================[ 2D Vector ]====*/


template<class T> 
class vec2
{
	public:

		/* Members */
		union {
			struct {
				union {
					T		x;
					T		u;
				};
				union {
					T		y;
					T		v;
				};
			};

			T				s[2];
		};

		/* Constructors */
		vec2();

		vec2(
				const vec2<T>	&v
			);

		vec2(
				T			nx,
				T			ny
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const vec2<T>	&v
			);


		/* Comparison */
		bool operator==(
				const vec2<T>	&v
			);

		bool operator!=(
				const vec2<T>	&v
			);


		/* Addition */
		vec2<T> operator+(
				const vec2<T>	&v
			);

		void operator+=(
				const vec2<T>	&v
			);

		vec2<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		vec2<T> operator+();


		/* Subtraction */
		vec2<T> operator-(
				const vec2<T>	&v
			);

		void operator-=(
				const vec2<T>	&v
			);

		vec2<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		vec2<T> operator-();


		/* Scalar/element-wise multiplication */
		vec2<T> operator*(
				const vec2<T>	&v
			);

		void operator*=(
				const vec2<T>	&v
			);

		vec2<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);


		/* Dot product */
		T dot(
				const vec2<T>	&v
			);

		T operator^(
				const vec2<T>	&v
			);


		/* Division */
		vec2<T> operator/(
				const vec2<T>	&v
			);

		void operator/=(
				const vec2<T>	&v
			);

		vec2<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Misc useful stuff */
		T minElement();

		T maxElement();

		void clamp(
				const T			min,
				const T			max
			);

		void saturate();

		void normalize();

		T length();

		T length2();

		vec2<T> reflection(
				const vec2<T>	&v
			);

		vec2<T> refraction(
				const vec2<T>	&v,
				const T			r
			);

};

/*===========================================================[ 3D Vector ]====*/

template<class T> 
class vec3
{
	public:

		/* Members */
		union {
			struct {
				union {
					T		x;
					T		u;
					T		r;
				};
				union {
					T		y;
					T		v;
					T		g;
				};
				union {
					T		z;
					T		w;
					T		b;
				};
			};

			T				s[3];
		};

		/* Constructors */
		vec3();

		vec3(
				const vec3<T>	&v
			);

		vec3(
				T			nx,
				T			ny,
				T			nz
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const vec3<T>	&v
			);


		/* Comparison */
		bool operator==(
				const vec3<T>	&v
			);

		bool operator!=(
				const vec3<T>	&v
			);


		/* Addition */
		vec3<T> operator+(
				const vec3<T>	&v
			);

		void operator+=(
				const vec3<T>	&v
			);

		vec3<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		vec3<T> operator+();


		/* Subtraction */
		vec3<T> operator-(
				const vec3<T>	&v
			);

		void operator-=(
				const vec3<T>	&v
			);

		vec3<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		vec3<T> operator-();


		/* Scalar/element-wise multiplication */
		vec3<T> operator*(
				const vec3<T>	&v
			);

		void operator*=(
				const vec3<T>	&v
			);

		vec3<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);


		/* Dot product */
		T dot(
				const vec3<T>	&v
			);

		T operator^(
				const vec3<T>	&v
			);


		/* Cross product */
		vec3<T> cross(
				const vec3<T>	&v
			);

		vec3<T> operator%(
				const vec3<T>	&v
			);


		/* Division */
		vec3<T> operator/(
				const vec3<T>	&v
			);

		void operator/=(
				const vec3<T>	&v
			);

		vec3<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Misc useful stuff */
		T minElement();

		T maxElement();

		void clamp(
				const T			min,
				const T			max
			);

		void saturate();

		void normalize();

		T length();

		T length2();

		vec3<T> reflection(
				const vec3<T>	&v
			);

		vec3<T> refraction(
				const vec3<T>	&v,
				const T			r
			);

};

/*===========================================================[ 4D Vector ]====*/

template<class T> 
class vec4
{
	public:

		/* Members */
		union {
			struct {
				union {
					T		x;
					T		r;
				};
				union {
					T		y;
					T		g;
				};
				union {
					T		z;
					T		b;
				};
				union {
					T		w;
					T		a;
				};
			};

			T				s[4];
		};

		/* Constructors */
		vec4();

		vec4(
				const vec4<T>	&v
			);

		vec4(
				T			nx,
				T			ny,
				T			nz,
				T			nw
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const vec4<T>	&v
			);


		/* Comparison */
		bool operator==(
				const vec4<T>	&v
			);

		bool operator!=(
				const vec4<T>	&v
			);


		/* Addition */
		vec4<T> operator+(
				const vec4<T>	&v
			);

		void operator+=(
				const vec4<T>	&v
			);

		vec4<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		vec4<T> operator+();


		/* Subtraction */
		vec4<T> operator-(
				const vec4<T>	&v
			);

		void operator-=(
				const vec4<T>	&v
			);

		vec4<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		vec4<T> operator-();


		/* Scalar/element-wise multiplication */
		vec4<T> operator*(
				const vec4<T>	&v
			);

		void operator*=(
				const vec4<T>	&v
			);

		vec4<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);


		/* Dot product */
		T dot(
				const vec4<T>	&v
			);

		T operator^(
				const vec4<T>	&v
			);


		/* Division */
		vec4<T> operator/(
				const vec4<T>	&v
			);

		void operator/=(
				const vec4<T>	&v
			);

		vec4<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Misc useful stuff */
		T minElement();

		T maxElement();

		void clamp(
				const T			min,
				const T			max
			);

		void saturate();

		void normalize();

		T length();

		T length2();

		vec4<T> reflection(
				const vec4<T>	&v
			);

		vec4<T> refraction(
				const vec4<T>	&v,
				const T			r
			);

};




/*=========================================================[ 2x2 Matrix ]====*/

template<class T>
class mat2
{
	public:
		friend class vec2<T>;

		/* Members */
		union {
			struct {
				T			a;
				T			c;

				T			b;
				T			d;
			};

			struct {
				vec2<T>		r[2];
			};

			T				s[4];
			T				t[2][2];
		};

	
		/* Constructors */
		mat2();

		mat2(
				const mat2<T>	&v
			);

		mat2(
				T				na,
				T				nb,
				T				nc,
				T				nd
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const mat2<T>	&v
			);


		/* Comparison */
		bool operator==(
				const mat2<T>	&v
			);

		bool operator!=(
				const mat2<T>	&v
			);

		bool isIdentity();


		/* Addition */
		mat2<T> operator+(
				const mat2<T>	&v
			);

		void operator+=(
				const mat2<T>	&v
			);

		mat2<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		mat2<T> operator+();


		/* Subtraction */
		mat2<T> operator-(
				const mat2<T>	&v
			);

		void operator-=(
				const mat2<T>	&v
			);

		mat2<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		mat2<T> operator-();


		/* Inner product */
		T iprod(
				const mat2<T>	&v,
				const u32		row,
				const u32		col
			);


		/* Multiplication */
		mat2<T> operator*(
				const mat2<T>	&v
			);

		void operator*=(
				const mat2<T>	&v
			);

		mat2<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);

		vec2<T> operator*(
				const vec2<T>	&v
			);


		/* Division */
		mat2<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Element-wise multiplication */
		mat2<T> operator&(
				const mat2<T>	&v
			);

		void operator&=(
				const mat2<T>	&v
			);


		/* Division with scalar */
		mat2<T> operator/(
				const mat2<T>	&v
			);

		void operator/=(
				const mat2<T>	&v
			);


		/* Misc matrix operations */
		void identity();

		void transpose();

		T det();

		void inv();

		T trace();



};

/*=========================================================[ 3x3 Matrix ]====*/

template<class T>
class mat3
{
	public:
		friend class vec3<T>;

		/* Members */
		union {
			struct {
				T			a;
				T			d;
				T			g;
				
				T			b;
				T			e;
				T			h;

				T			c;
				T			f;
				T			i;
			};





			T				s[9];
			T				t[3][3];
		};

	
		/* Constructors */
		mat3();

		mat3(
				const mat3<T>	&v
			);

		mat3(
				T				na,
				T				nb,
				T				nc,

				T				nd,
				T				ne,
				T				nf,

				T				ng,
				T				nh,
				T				ni
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const mat3<T>	&v
			);


		/* Comparison */
		bool operator==(
				const mat3<T>	&v
			);

		bool operator!=(
				const mat3<T>	&v
			);

		bool isIdentity();


		/* Addition */
		mat3<T> operator+(
				const mat3<T>	&v
			);

		void operator+=(
				const mat3<T>	&v
			);

		mat3<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		mat3<T> operator+();


		/* Subtraction */
		mat3<T> operator-(
				const mat3<T>	&v
			);

		void operator-=(
				const mat3<T>	&v
			);

		mat3<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		mat3<T> operator-();


		/* Inner product */
		T iprod(
				const mat3<T>	&v,
				const u32		row,
				const u32		col
			);


		/* Multiplication */
		mat3<T> operator*(
				const mat3<T>	&v
			);

		void operator*=(
				const mat3<T>	&v
			);

		mat3<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);

		vec3<T> operator*(
				const vec3<T>	&v
			);


		/* Division */
		mat3<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Element-wise multiplication */
		mat3<T> operator&(
				const mat3<T>	&v
			);

		void operator&=(
				const mat3<T>	&v
			);


		/* Division with scalar */
		mat3<T> operator/(
				const mat3<T>	&v
			);

		void operator/=(
				const mat3<T>	&v
			);


		/* Misc matrix operations */
		void identity();

		void transpose();

		T det();

		void inv();

		T trace();



};

/*=========================================================[ 4x4 Matrix ]====*/

template<class T>
class mat4
{
	public:
		friend class vec4<T>;

		/* Members */
		union {
			struct {
				T			a;
				T			e;
				T			i;
				T			m;

				T			b;
				T			f;
				T			j;
				T			n;

				T			c;
				T			g;
				T			k;
				T			o;

				T			d;
				T			h;
				T			l;
				T			p;
			};
			struct {
				T			R00;
				T			R01;
				T			R02;
				T			N0;

				T			R10;
				T			R11;
				T			R12;
				T			N1;

				T			R20;
				T			R21;
				T			R22;
				T			N2;

				T			T0;
				T			T1;
				T			T2;
				T			N3;
			};





			T				s[16];
			T				t[4][4];
		};

	
		/* Constructors */
		mat4();

		mat4(
				const mat4<T>	&v
			);

		mat4(
				T				na,
				T				nb,
				T				nc,
				T				nd,

				T				ne,
				T				nf,
				T				ng,
				T				nh,

				T				ni,
				T				nj,
				T				nk,
				T				nl,

				T				nm,
				T				nn,
				T				no,
				T				np
			);

		/* Functions */

		/* Assignment */
		void operator=(
				const mat4<T>	&v
			);


		/* Comparison */
		bool operator==(
				const mat4<T>	&v
			);

		bool operator!=(
				const mat4<T>	&v
			);

		bool isIdentity();


		/* Addition */
		mat4<T> operator+(
				const mat4<T>	&v
			);

		void operator+=(
				const mat4<T>	&v
			);

		mat4<T> operator+(
				const T			v
			);

		void operator+=(
				const T			v
			);

		mat4<T> operator+();


		/* Subtraction */
		mat4<T> operator-(
				const mat4<T>	&v
			);

		void operator-=(
				const mat4<T>	&v
			);

		mat4<T> operator-(
				const T			v
			);

		void operator-=(
				const T			v
			);

		mat4<T> operator-();


		/* Inner product */
		T iprod(
				const mat4<T>	&v,
				const u32		row,
				const u32		col
			);


		/* Multiplication */
		mat4<T> operator*(
				const mat4<T>	&v
			);

		void operator*=(
				const mat4<T>	&v
			);

		mat4<T> operator*(
				const T			v
			);

		void operator*=(
				const T			v
			);

		vec4<T> operator*(
				const vec4<T>	&v
			);


		/* Division */
		mat4<T> operator/(
				const T			v
			);

		void operator/=(
				const T			v
			);


		/* Element-wise multiplication */
		mat4<T> operator&(
				const mat4<T>	&v
			);

		void operator&=(
				const mat4<T>	&v
			);


		/* Division with scalar */
		mat4<T> operator/(
				const mat4<T>	&v
			);

		void operator/=(
				const mat4<T>	&v
			);


		/* Misc matrix operations */
		void identity();

		void transpose();

		T det();

		void inv();

		T trace();

		void print();

	private:
		T det2(
				T n0,
				T n1,
				T n2,
				T n3
			);

};

/*==================================================[ Template uglyness ]====*/

typedef vec2<float>			vec2f;
typedef vec2<double>		vec2d;

typedef vec3<float>			vec3f;
typedef vec3<double>		vec3d;

typedef vec4<float>			vec4f;
typedef vec4<double>		vec4d;

typedef vec3<float>			rgbf;
typedef vec4<float>			rgbaf;


typedef mat2<float>			mat2f;
typedef mat2<double>		mat2d;

typedef mat3<float>			mat3f;
typedef mat3<double>		mat3d;

typedef mat4<float>			mat4f;
typedef mat4<double>		mat4d;


#define RC_VMATH
#include "VectorMath.cpp"


#endif /* __RC_VECTORMATH_H__ */

