#include "RenderChimp.h"
#include "VertexArray.h"

/* Warnings */
static const Result_t WRN_HAS_NORMALS = {
	RESULT_WARNING, "Vertex array already has normals" };

static const Result_t WRN_NO_POS = {
	RESULT_WARNING, "No vertex positions found" };

static const Result_t WRN_TOO_MANY_ATTR = {
	RESULT_WARNING, "Too many attributes" };

static const Result_t WRN_OFFSET = {
	RESULT_WARNING, "The offset is greater or equal to stride" };

static const Result_t WRN_ATTR_EXISTS = {
	RESULT_WARNING, "The attribute is already set" };

/*---------------------------------------------------------------------------*/

VertexArray::VertexArray(
		const char				*nidentifier,

		f32						*new_array,
		u32						nstride,
		u32						n,

		u32						purge_level
	) : Resource(nidentifier, purge_level, RESOURCE_VERTEX_ARRAY)
{
	if (strlen(nidentifier) == 0)
		REPORT(ERR_ARG);

	n_attribs = 0;

	n_elements = n;
	stride = nstride;

	vertexArray = (f32 *) malloc(sizeof(f32) * n * stride);
	memcpy(vertexArray, new_array, sizeof(f32) * n * stride);

	size = sizeof(f32) * n * stride;
}

/*---------------------------------------------------------------------------*/

VertexArray::~VertexArray()
{
	u32 i;

	if (vertexArray)
		free(vertexArray);

	for (i = 0; i < n_attribs; i++)
		free(attrib_name[i]);
}

/*---------------------------------------------------------------------------*/

Resource *VertexArray::clone(
		const char				*clone_name,

		u32						purge_level
	)
{
	u32 i;

	VertexArray *new_res = new VertexArray(identifier, vertexArray, 
		stride, n_elements, purge_level);

	for (i = 0; i < n_attribs; i++) {
		new_res->setAttribute(attrib_offset[i], attrib_length[i], attrib_name[i]);
	}

	return new_res;
}

/*---------------------------------------------------------------------------*/

void VertexArray::setAttribute(
		u32						offset,
		u32						length,
		const char				*name
	)
{
	i32 i;

	if (length == 0)
		return;

	if (n_attribs >= MAX_ATTRIBS) {
		REPORT(WRN_TOO_MANY_ATTR);
		return;
	}

	if (offset >= stride) {
		REPORT(WRN_OFFSET);
		return;
	}

	for (i = 0; i < (i32) n_attribs; i++) {
		if (strcmp(name, attrib_name[i]) == 0) {
			/* The attribute already exists */

			if (attrib_offset[i] != offset || attrib_length[i] != length) {
				REPORT(WRN_ATTR_EXISTS);
			}
			return;
		}
	}

	i = n_attribs - 1;

	while (i >= 0 && offset < attrib_offset[i]) {
		attrib_offset[i + 1] = attrib_offset[i];
		attrib_length[i + 1] = attrib_length[i];
		attrib_name[i + 1] = attrib_name[i];
		i--;
	}

	i++;

	attrib_offset[i] = offset;
	attrib_length[i] = length;
	attrib_name[i] = duplicateString(name);

	n_attribs++;
}

/*---------------------------------------------------------------------------*/

f32 *VertexArray::getArray()
{
	return vertexArray;
}

/*---------------------------------------------------------------------------*/

u32 VertexArray::getLength()
{
	return n_elements;
}

/*---------------------------------------------------------------------------*/

void VertexArray::pushToShader(
		ShaderProgram			*shader
	)
{
	u32 i;

	/* Set up attributes */
	if (n_attribs == 1) {
		shader->setAttributePointer(attrib_name[0],
			attrib_length[0],
			false,
			0,
			vertexArray);
	} else {
		for (i = 0; i < n_attribs; i++) {

			shader->setAttributePointer(attrib_name[i],
				attrib_length[i],
				false,
				stride,
				&vertexArray[attrib_offset[i]]);

		}
	}
}

/*---------------------------------------------------------------------------*/

void VertexArray::generateNormals(
		IndexArray			*indexArray
	)
{
	u32 i, offset;
	u32 old_stride = stride;
	u32 n_indices;
	f32 *new_array;
	f32 *old_iter, *new_iter;
	u16 *index_array;

	for (i = 0; i < n_attribs; i++) {

		if (strcmp(attrib_name[i], "Vertex") == 0)
			break;

	}

	if (i == n_attribs) {
		REPORT(WRN_NO_POS);
		return;
	}

	offset = attrib_offset[i];

	for (i = 0; i < n_attribs; i++) {
		if (strcmp(attrib_name[i], "Normal") == 0) {
			REPORT(WRN_HAS_NORMALS);
			return;
		}
	}

	stride += 3;

	setAttribute(stride - 3, 3, "Normal");

	new_array = (f32 *) malloc(sizeof(f32) * n_elements * stride);

	old_iter = vertexArray;
	new_iter = new_array;

	for (i = 0; i < n_elements; i++) {
		memcpy(new_iter, old_iter, old_stride * sizeof(f32));
		new_iter += old_stride;

		new_iter[0] = 0.0f;
		new_iter[1] = 0.0f;
		new_iter[2] = 0.0f;

		new_iter += 3;

		old_iter += old_stride;
	}

	n_indices = indexArray->getLength();
	index_array = indexArray->getArray();

	for (i = 0; i < n_indices; i += 3) {
		f32 *p0 = &vertexArray[index_array[i  ] * old_stride + offset];
		f32 *p1 = &vertexArray[index_array[i+1] * old_stride + offset];
		f32 *p2 = &vertexArray[index_array[i+2] * old_stride + offset];

		f32 *t0 = &new_array[index_array[i  ] * stride + old_stride];
		f32 *t1 = &new_array[index_array[i+1] * stride + old_stride];
		f32 *t2 = &new_array[index_array[i+2] * stride + old_stride];

		vec3f v0 = vec3f(p0[0], p0[1], p0[2]);
		vec3f v1 = vec3f(p1[0], p1[1], p1[2]);
		vec3f v2 = vec3f(p2[0], p2[1], p2[2]);

		vec3f e0 = v2 - v1;
		vec3f e1 = v1 - v0;

		vec3f normal = e0 % e1;

		f32 area = 0.5f * normal.length();

		normal.normalize();

		normal *= area;

		t0[0] += normal.x;
		t0[1] += normal.y;
		t0[2] += normal.z;

		t1[0] += normal.x;
		t1[1] += normal.y;
		t1[2] += normal.z;

		t2[0] += normal.x;
		t2[1] += normal.y;
		t2[2] += normal.z;
	}

	new_iter = &new_array[old_stride];

	/* Normalize our new normals */
	for (i = 0; i < n_elements; i++) {
		f32 *p = new_iter;

		f32 len = p[0] * p[0] + p[1] * p[1] + p[2] * p[2];

		if (len != 0.0f) {
			f32 invlen = 1.0f / sqrt(len);

			p[0] *= invlen;
			p[1] *= invlen;
			p[2] *= invlen;
		} else { /* Just in case... */
			p[0] = 1.0f;
			p[1] = 0.0f;
			p[2] = 0.0f;
		}

		new_iter += stride;
	}

	free(vertexArray);
	vertexArray = new_array;

}

/*---------------------------------------------------------------------------*/

void VertexArray::setVertexArray(
		VertexArray				**target_vertex_array,
		VertexArray				*new_vertex_array,

		bool					instantiate_resource
	)
{
	if (*target_vertex_array)
		sceneGraph.dereferenceResource(*target_vertex_array);

	if (instantiate_resource) {
		*target_vertex_array = (VertexArray *) new_vertex_array->instantiate();
	} else {
		*target_vertex_array = new_vertex_array;
	}
}

/*---------------------------------------------------------------------------*/

#ifdef DEBUG

void VertexArray::print(
		FILE		*f
	)
{

#ifdef LOG_TO_FILE
	#define prnt0(a)		fprintf(f, a)
	#define prnt1(a, b)		fprintf(f, a, b)
#else /* LOG_TO_FILE */
	#define prnt0(a)		printf(a)
	#define prnt1(a, b)		printf(a, b)
#endif /* LOG_TO_FILE */

	prnt1("    [ N ELEMS  ][ %u ]\n", n_elements);
	prnt1("    [ STRIDE   ][ %u ]\n", stride);
	prnt1("    [ N ATTR.  ][ %u ]\n", n_attribs);

}

#endif /* DEBUG */

/*---------------------------------------------------------------------------*/

u32 VertexArray::getNAttributes()
{
	return n_attribs;
}

/*---------------------------------------------------------------------------*/

char *VertexArray::getAttributeName(
		u32						i
	)
{
	return attrib_name[i];
}

/*---------------------------------------------------------------------------*/

