#include <cstdio>
#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Node.h"
#include "Transformable.h"
#include "Camera.h"
#include "World.h"

/* Warnings */
static const Result_t WRN_NO_CAM = {
	RESULT_WARNING, "No active camera" };

/*---------------------------------------------------------------------------*/

World::World(
		const char	*nname
	) : Node(nname, NODE_WORLD)
{
	drawInfo.matTransform = new MatrixTransformation;
	drawInfo.activeShader = 0;

	clearColorBuffer = true;
	clearDepthBuffer = true;
	clearStencilBuffer = false;

	clearColor = rgbaf(0.5f, 0.5f, 0.5f, 1.0f);
	clearDepth = 1.0f;
	clearStencil = 0;

	standardState.clear();

	standardState.enableCulling(CULL_BACK);
	standardState.enableDepth(CMP_LESS);
	standardState.enableBlend(SB_SRC_ALPHA, DB_ONE_MINUS_SRC_ALPHA);
	standardState.disableStencil();
	standardState.disableScissor();

	activeCamera = 0;

	full_viewport = true;

	RenderState::setState(&standardState.settings, 0xFF, true);
}

/*---------------------------------------------------------------------------*/

World::~World()
{
	delete drawInfo.matTransform;
}

/*---------------------------------------------------------------------------*/

void World::drawAll(
		DeviceState_t	*device_state
	)
{
	u32 flags = 0;
	mat4f viewMatrix;
	mat4f worldMatrix;
	mat4f viewProjMatrix;
	mat4f projectionMatrix;
	vec3f viewPos;

	if (full_viewport) {
		u32 dw, dh;
		apeGetDisplaySize(&dw, &dh);
		apeSetViewport(0, 0, dw, dh);
		sceneGraph.resize(dw, dh);
	} else {
		apeSetViewport(viewport_x, viewport_y, viewport_width, viewport_height);
		sceneGraph.resize(viewport_width, viewport_height);
	}

	if (clearColorBuffer) {
		glClearColor(clearColor.r, clearColor.g, clearColor.b, clearColor.a);
		flags |= GL_COLOR_BUFFER_BIT;
	}

	if (clearDepthBuffer) {
		glClearDepthf(clearDepth);
		flags |= GL_DEPTH_BUFFER_BIT;
	}

	if (clearStencilBuffer) {
		glClearStencil(clearStencil);
		flags |= GL_STENCIL_BUFFER_BIT;
	}

	glClear(flags);

	GL_ASSERT();

	if (!activeCamera) {
		REPORT(WRN_NO_CAM);
		return;
	}

	GL_ASSERT();

	RenderState::setState(&standardState.settings, 0xFF, true);
	drawInfo.currentState = standardState.settings;

	GL_ASSERT();

	worldMatrix.identity();
	drawInfo.matTransform->world = &worldMatrix;

	update(&drawInfo);

	activeCamera->calculateViewTransform(&viewMatrix);

	viewPos = vec3f(activeCamera->world.T0, activeCamera->world.T1, activeCamera->world.T2);
	viewPos = -viewPos;

	projectionMatrix = activeCamera->projectionMatrix;
	viewProjMatrix = projectionMatrix * viewMatrix;

	drawInfo.matTransform->prepareFrame(&viewMatrix, &projectionMatrix,
		&viewProjMatrix, viewPos, activeCamera->projectionType);
	drawInfo.activeShader = 0;
	drawInfo.deviceState = device_state;

	draw(&drawInfo);

	sceneGraph.text->printAll();
}

/*---------------------------------------------------------------------------*/

void World::transformCamera(
		mat4f *matrix
	)
{
	/* Hooray! Camera transform is done! */
}

/*---------------------------------------------------------------------------*/

void World::setActiveCamera(
		Camera *camera
	)
{
	activeCamera = camera;
}

/*---------------------------------------------------------------------------*/

Camera *World::getActiveCamera()
{
	return activeCamera;
}

/*---------------------------------------------------------------------------*/

void World::setClearColorBuffer(
		bool		do_clear,
		rgbaf		color
	)
{
	clearColorBuffer = do_clear;
	clearColor = color;
}

/*---------------------------------------------------------------------------*/

void World::setClearDepthBuffer(
		bool		do_clear,
		f32			depth
	)
{
	clearDepthBuffer = do_clear;
	clearDepth = depth;
}

/*---------------------------------------------------------------------------*/

void World::setClearStencilBuffer(
		bool		do_clear,
		i32			stencil
	)
{
	clearStencilBuffer = do_clear;
	clearStencil = stencil;
}

/*---------------------------------------------------------------------------*/

Node *World::duplicateSelf()
{
	World *c = sceneGraph.createWorld(name);

	c->setActiveCamera(activeCamera);
	c->setClearColorBuffer(clearColorBuffer, clearColor);
	c->setClearDepthBuffer(clearDepthBuffer, clearDepth);
	c->setClearStencilBuffer(clearStencilBuffer, clearStencil);

	return c;
}

/*---------------------------------------------------------------------------*/

RenderState *World::getStandardState()
{
	return &standardState;
}

/*---------------------------------------------------------------------------*/

void World::setViewport(
		i32			x,
		i32			y,
		u32			width,
		u32			height
	)
{
	viewport_x = x;
	viewport_y = y;
	viewport_width = width;
	viewport_height = height;

	full_viewport = false;
}

/*---------------------------------------------------------------------------*/

void World::setFullViewport()
{
	full_viewport = true;
}

/*---------------------------------------------------------------------------*/

